﻿using System.Linq;
using Styx.Combat.CombatRoutine;
using Styx.Logic.Pathing;
using Styx.WoWInternals;
using TreeSharp;
using Action = TreeSharp.Action;
using Druid_ID = Hera.ClassHelpers.Druid.IDs;

namespace Hera
{
    public partial class Fpsware
    {
        #region Pull
        private Composite _pullBehavior;
        public override Composite PullBehavior
        {
            get { if (_pullBehavior == null) { Utils.Log("Creating 'Pull' behavior"); _pullBehavior = CreatePullBehavior(); }  return _pullBehavior; }
        }

        private PrioritySelector CreatePullBehavior()
        {
            return new PrioritySelector(

                // If we can't navigate to the target blacklist it.
                new NeedToNavigatePath(new NavigatePath()),

                // Check if the target is suitable for pulling, if not blacklist it
                new NeedToBlacklistPullTarget(new BlacklistPullTarget()),

                // Check pull timers and blacklist bad pulls where required
                new NeedToCheckPullTimer(new BlacklistPullTarget()),



                // Shapeshift Bear
                new NeedToShapeshiftBear(new ShapeshiftBear()),

                // Shapeshift Moonkin
                new NeedToShapeshiftMoonkin(new ShapeshiftMoonkin()),

                // Shapeshift Cat
                new NeedToShapeshiftCat(new ShapeshiftCat()),

                // Pull dependant upon your SPEC. Not your currently shapeshifted form
                new Switch<ClassHelpers.Druid.ClassType>(ret => ClassHelpers.Druid.ClassSpec,
                                                new SwitchArgument<ClassHelpers.Druid.ClassType>(BalancePull, ClassHelpers.Druid.ClassType.Balance),
                                                new SwitchArgument<ClassHelpers.Druid.ClassType>(BalancePull, ClassHelpers.Druid.ClassType.Restoration),
                                                new SwitchArgument<ClassHelpers.Druid.ClassType>(BalancePull, ClassHelpers.Druid.ClassType.Untalented),
                                                new SwitchArgument<ClassHelpers.Druid.ClassType>(FeralPull, ClassHelpers.Druid.ClassType.Feral)


                )
                );
        }


        #region Pull Balance
        private Composite BalancePull
        {
            get
            {
                return new PrioritySelector(

                    // Casting so stop moving
                    new NeedToCastingSoStopMoving(new CastingSoStopMoving()),

                    // Distance Check Balance
                    new NeedToDistanceCheckBalance(new DistanceCheckBalance()),

                    // Pull Spell Balance
                    new NeedToPullSpellBalance(new PullSpellBalance()),

                    // Wrath Pull - Fall through
                    new NeedToWrathPull(new WrathPull())

                    );
            }
        }
        #endregion

        #region Pull Cat / Bear
        private Composite FeralPull
        {
            get
            {
                return new PrioritySelector(

                    // Distance Check Feral
                    new NeedToDistanceCheckFeral(new DistanceCheckFeral()),

                    // Face - only if we're not moving
                    new NeedToFacePull(new FacePull()),

                    // Feral Pull Automatic - Prowl if its a caster or FFF if its melee.
                    // Well, thats the idea anyway :)
                    new NeedToFeralPullAutomatic(new FeralPullAutomatic()),

                    // Auto Attack Low Level Pull
                    new NeedToAutoAttackLowLevelPull(new AutoAttackLowLevelPull()),

                    // Feral Charge Cat
                    new NeedToFeralChargeCat(new FeralChargeCat()),

                    // Dash
                    new NeedToDash(new Dash()),

                    // Ravage
                    new NeedToRavage(new Ravage()),

                    // Pounce
                    new NeedToPounce(new Pounce()),

                    // Faerie Fire Feral Pull
                    new NeedToFaerieFireFeralPull(new FaerieFireFeralPull()),

                    // Prowl
                    new NeedToProwl(new Prowl()),

                    // Distance Check Feral
                    new NeedToDistanceCheckFeral(new DistanceCheckFeral())

                    );
            }
        }
        #endregion
        #endregion

        #region Pull Timer / Timeout
        public class NeedToCheckPullTimer : Decorator
        {
            public NeedToCheckPullTimer(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (Settings.LazyRaider.Contains("always")) return false;

                return Target.PullTimerExpired;
            }
        }

        public class CheckPullTimer : Action
        {
            protected override RunStatus Run(object context)
            {
                Utils.Log(string.Format("Unable to pull {0}, blacklisting and finding another target.", Me.CurrentTarget.Name), System.Drawing.Color.FromName("Red"));
                Target.BlackList(120);
                Me.ClearTarget();

                return RunStatus.Success;
            }
        }
        #endregion

        #region Combat Timer / Timeout
        public class NeedToCheckCombatTimer : Decorator
        {
            public NeedToCheckCombatTimer(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (Utils.IsBattleground) return false;
                if (!Me.GotTarget || Me.CurrentTarget.Dead) return false;
                if (Target.IsElite) return false;
                if (CT.Name.Contains("Training Dummy")) return false;
                if (Me.IsInInstance) return false;
                if (ClassHelpers.Druid.ClassSpec == ClassHelpers.Druid.ClassType.Restoration) return false;
                
                return Target.CombatTimerExpired && Target.IsHealthPercentAbove(98);
            }
        }

        public class CheckCombatTimer : Action
        {
            protected override RunStatus Run(object context)
            {
                Utils.Log(string.Format("Combat with {0} is bugged, blacklisting and finding another target.", Me.CurrentTarget.Name), System.Drawing.Color.FromName("Red"));
                Target.BlackList(60);
                Utils.LagSleep();

                return RunStatus.Success;
            }
        }
        #endregion


        #region Behaviours

        #region Common

        #region Auto Attack During Pull
        public class NeedToAutoAttackPull : Decorator
        {
            public NeedToAutoAttackPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (!Me.GotTarget) return false;
                if (Target.IsWithinInteractRange && !Me.IsAutoAttacking) return true;

                return false;
            }
        }

        public class AutoAttackPull : Action
        {
            protected override RunStatus Run(object context)
            {
                Utils.AutoAttack(true);

                return RunStatus.Failure;
            }
        }
        #endregion

        #region Navigate Path
        public class NeedToNavigatePath : Decorator
        {
            public NeedToNavigatePath(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (!Me.GotTarget) return false;
                if (Settings.LazyRaider.Contains("always")) return false;
                
                return !Navigator.CanNavigateFully(Me.Location,CT.Location,20);
            }
        }

        public class NavigatePath : Action
        {
            protected override RunStatus Run(object context)
            {
                Utils.Log("Can not navigate to target's location. Blacklisting",Utils.Colour("Red"));
                Target.BlackList(Utils.IsBattleground ? 10 : 30);

                return RunStatus.Success;
            }
        }
        #endregion

        #region Pull Spell
        public class NeedToPullSpell: Decorator
        {
            public NeedToPullSpell(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                string dpsSpell = Settings.PullSpellBalance;
                if (!Timers.Expired("PullSpellCast",5000)) return false;

                if (Me.Combat) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (!Settings.LazyRaider.Contains("always"))
                {
                    if (Target.IsDistanceMoreThan(Spell.MaxDistance(Settings.PullSpellBalance)))
                    {
                        Movement.MoveTo((float)Spell.MaxDistance(Settings.PullSpellBalance) - 2);
                        System.Threading.Thread.Sleep(1500);
                        return false;
                    }

                    if (Utils.IsInLineOfSight(CT.Location) && Me.IsMoving)
                    {
                        Movement.StopMoving();
                        Utils.LagSleep();
                    }
                }
                if (Me.IsMoving) return false;

                return (Spell.CanCast(dpsSpell));
            }
        }

        public class PullSpell : Action
        {
            protected override RunStatus Run(object context)
            {
                string dpsSpell = Settings.PullSpellBalance;

                Target.Face();
                Utils.LagSleep();


                bool result = Spell.Cast(dpsSpell);
                Utils.LagSleep();
                Utils.WaitWhileCasting();

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Icy Touch Pull
        public class NeedToIcyTouchPull : Decorator
        {
            public NeedToIcyTouchPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Icy Touch";

                //if (Spell.IsKnown(Settings.PullSpell) && !Spell.IsOnCooldown(Settings.PullSpell)) return false;
                if (Target.IsDistanceMoreThan(Spell.MaxDistance(dpsSpell)) && Spell.CanCast("Death Grip")) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (Target.IsDistanceMoreThan(Spell.MaxDistance(dpsSpell))) return false;
		
                //return (Spell.CanCast(dpsSpell));
                return (Spell.IsKnown(dpsSpell) && !Spell.IsOnCooldown(dpsSpell));
            }
        }

        public class IcyTouchPull : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Icy Touch";
                bool result = Spell.Cast(dpsSpell);

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Death Grip Pull
        public class NeedToDeathGripPull : Decorator
        {
            public NeedToDeathGripPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Death Grip";

                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (Target.IsDistanceMoreThan(Spell.MaxDistance(dpsSpell))) return false;
                if (Target.IsDistanceLessThan(Spell.MaxDistance("Icy Touch")) && Spell.CanCast("Icy Touch")) return false;
                if (Me.UnholyRuneCount < 1) return false;
		
                //return (Spell.CanCast(dpsSpell));
                return (Spell.IsKnown(dpsSpell) && !Spell.IsOnCooldown(dpsSpell));
            }
        }

        public class DeathGripPull : Action
        {
            protected override RunStatus Run(object context)
            {
                if (Me.IsMoving) WoWMovement.MoveStop();
                System.Threading.Thread.Sleep(500);
                Target.Face();

                const string dpsSpell = "Death Grip";
                bool result = Spell.Cast(dpsSpell);


                System.Threading.Thread.Sleep(1000);


                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Death Coil
        public class NeedToDeathCoilPull : Decorator
        {
            public NeedToDeathCoilPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Death Coil";

                if (Me.CurrentRunicPower < Spell.PowerCost(dpsSpell)) return false;
                //if (Me.RunicPowerPercent < Settings.DeathCoil) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (Target.IsDistanceMoreThan(Spell.MaxDistance(dpsSpell))) return false;
		
                return (Spell.CanCast(dpsSpell));
            }
        }

        public class DeathCoilPull : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Death Coil";
                bool result = Spell.Cast(dpsSpell);
                

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        




        #region Face Pull
        public class NeedToFacePull : Decorator
        {
            public NeedToFacePull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (Settings.LazyRaider.Contains("always")) return false;
                return (!Me.IsMoving && Me.GotTarget);
            }
        }

        public class FacePull : Action
        {
            protected override RunStatus Run(object context)
            {
                Target.Face();
                return RunStatus.Failure;
            }
        }
        #endregion


        #endregion

        #region Feral Cat

        #region Prowl
        public class NeedToProwl : Decorator
        {
            public NeedToProwl(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Prowl";

                if (!Settings.PullSpellFeralCat.ToUpper().Contains("PROWL")) return false;
                if (!ClassHelpers.Druid.Shapeshift.IsCatForm) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (Self.IsBuffOnMe(dpsSpell)) return false;
                if (Target.IsDistanceMoreThan(Settings.StealthPullDistance)) return false;
                if (!Spell.IsKnown("Pounce")) return false;

                return (Spell.CanCast(dpsSpell));
            }
        }

        public class Prowl : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Prowl";
                bool result = Spell.Cast(dpsSpell);

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Faerie Fire Feral Pull
        public class NeedToFaerieFireFeralPull : Decorator
        {
            public NeedToFaerieFireFeralPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Faerie Fire (Feral)";

                if (!Me.GotTarget) return false;
                //if (!Settings.PullSpellFeralCat.ToUpper().Contains("FAERIE FIRE")) return false;
                if (Settings.PullSpellFeralCat.ToUpper().Contains("PROWL")) return false;
                if (Self.IsBuffOnMe(Druid_ID.Prowl)) return false;
                if (Settings.FeralCatRavage.Contains("always") && Spell.IsKnown("Feral Charge (Cat)") && !Spell.IsOnCooldown("Feral Charge (Cat)") && Target.Distance > 8) return false;
                if (!ClassHelpers.Druid.Shapeshift.IsCatForm && !ClassHelpers.Druid.Shapeshift.IsBearForm) return false;
                if (CT.Auras.ContainsKey(dpsSpell)) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (Target.Distance > Spell.MaxDistance(dpsSpell)) return false;

                return (Spell.CanCast(dpsSpell));
            }
        }

        public class FaerieFireFeralPull : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Faerie Fire (Feral)";
                bool result = Spell.Cast(dpsSpell);

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Pounce
        public class NeedToPounce : Decorator
        {
            public NeedToPounce(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Pounce";

                if (!ClassHelpers.Druid.Shapeshift.IsCatForm) return false;
                if (!Self.IsBuffOnMe(Druid_ID.Prowl)) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (CT.Distance > CT.InteractRange) return false;
                
                return (Spell.CanCast(dpsSpell));
            }
        }

        public class Pounce : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Pounce";
                bool result = Spell.Cast(dpsSpell);

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Dash
        public class NeedToDash : Decorator
        {
            public NeedToDash(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Dash";

                if (!ClassHelpers.Druid.Shapeshift.IsCatForm) return false;
                if (!Self.IsBuffOnMe(Druid_ID.Prowl)) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (Target.IsDistanceLessThan(25)) return false;
                
                return (Spell.CanCast(dpsSpell));
            }
        }

        public class Dash : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Dash";
                bool result = Spell.Cast(dpsSpell);

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Ravage
        public class NeedToRavage : Decorator
        {
            public NeedToRavage(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Ravage";

                if (!ClassHelpers.Druid.Shapeshift.IsCatForm) return false;
                if (!Self.IsBuffOnMe(Druid_ID.Prowl)) return false;
                if (!Me.IsSafelyBehind(CT)) return false;
                if (Self.Immobilised) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;
                if (CT.Distance > CT.InteractRange) return false;

                return (Spell.CanCast(dpsSpell));
            }
        }

        public class Ravage : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Ravage";
                bool result = Spell.Cast(dpsSpell);

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Feral Pull Automatic
        public class NeedToFeralPullAutomatic : Decorator
        {
            public NeedToFeralPullAutomatic(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (!ClassHelpers.Druid.Shapeshift.IsCatForm) return false;
                if (Self.IsBuffOnMe(Druid_ID.Prowl)) return false;
                if (!Settings.PullSpellFeralCat.ToUpper().Contains("AUTOMATIC")) return false;

                bool rangedMob = Utils.RangedCapableMobs.Any(mobEntry => CT.Entry == mobEntry);

                if (!rangedMob)
                    if (Me.GotTarget && !CT.IsPlayer && Me.CurrentTarget.Class == WoWClass.Warrior) return false;
                if (Spell.IsKnown("Prowl") && !Spell.IsOnCooldown("Prowl")) Spell.Cast("Prowl");

                return false;
            }
        }

        public class FeralPullAutomatic : Action
        {
            protected override RunStatus Run(object context)
            {
              
                return RunStatus.Failure;
            }
        }
        #endregion

        #region Feral Charge Cat
        public class NeedToFeralChargeCat : Decorator
        {
            public NeedToFeralChargeCat(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Feral Charge (Cat)";

                if (!ClassHelpers.Druid.Shapeshift.IsCatForm) return false;
                if (Target.IsDistanceMoreThan(25)) return false;
                if (Target.IsDistanceLessThan(8)) return false;
                if (!Settings.FeralCatRavage.Contains("always"))
                    if (!Self.IsBuffOnMe(Druid_ID.Prowl)) return false;
                if (Self.IsBuffOnMe(Druid_ID.Dash)) return false;
                if (Self.Immobilised) return false;
                if (!Utils.CombatCheckOk(dpsSpell, false)) return false;

                return (Spell.CanCast(dpsSpell));
            }
        }

        public class FeralChargeCat : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Feral Charge (Cat)";
                bool result = Spell.Cast(dpsSpell);
                Timers.Reset("FeralCharge");

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Auto Attack Low Level Pull
        public class NeedToAutoAttackLowLevelPull : Decorator
        {
            public NeedToAutoAttackLowLevelPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                return (!Spell.IsKnown("Pounce") && !Me.IsAutoAttacking);
            }
        }

        public class AutoAttackLowLevelPull : Action
        {
            protected override RunStatus Run(object context)
            {
                Utils.AutoAttack(true);
                return RunStatus.Failure;
            }
        }
        #endregion

        #endregion

        #region Feral Bear

        #endregion

        #region Balance

        #region Distance Check Balance
        public class NeedToDistanceCheckBalance : Decorator
        {
            public NeedToDistanceCheckBalance(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                if (Settings.LazyRaider.ToUpper().Contains("ALWAYS")) return false;
                if (!Me.GotTarget || Me.CurrentTarget.Dead) return false;
                string pullSpell = "Wrath";
                if (ClassHelpers.Druid.CLCPullSpellBalance.ToUpper().Contains("AUTOMATIC")) pullSpell = ClassHelpers.Druid.PullSpellBalance;

                double minDistance = Spell.MaxDistance(pullSpell) - 7;
                double maxDistance = Spell.MaxDistance(pullSpell) - 2;
                
                if (Me.IsCasting) return false;
                if (!Target.IsFleeing)
                    if (!Me.CurrentTarget.IsMoving && !Me.IsMoving && Target.IsDistanceLessThan(Spell.MaxDistance(pullSpell))) return false;
                if (Target.IsDistanceLessThan(minDistance) && Me.IsMoving) {WoWMovement.MoveStop(); return false;}
                if (Target.IsDistanceMoreThan(maxDistance)) return true;

                return false;
            }
        }

        public class DistanceCheckBalance : Action
        {
            protected override RunStatus Run(object context)
            {
                string pullSpell = "Wrath";
                if (ClassHelpers.Druid.CLCPullSpellBalance.ToUpper().Contains("AUTOMATIC")) pullSpell = ClassHelpers.Druid.PullSpellBalance;

                float distanceMoveTo = (float)Spell.MaxDistance(pullSpell) - 6;

                Movement.MoveTo(distanceMoveTo);
                Timers.Reset("Pull");

                return RunStatus.Success;
            }
        }
        #endregion

        #region Pull Spell Balance
        public class NeedToPullSpellBalance : Decorator
        {
            public NeedToPullSpellBalance(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                string dpsSpell = Settings.PullSpellBalance;

                if (!Me.CurrentTarget.InLineOfSightOCD) return false;
                if (Me.IsMoving) return false;
                if (!Timers.Expired("Pull",2000)) return false;

                if (ClassHelpers.Druid.CLCPullSpellBalance.ToUpper().Contains("AUTOMATIC")) dpsSpell = ClassHelpers.Druid.PullSpellBalance;

                return Utils.CombatCheckOk(dpsSpell, false) && (Spell.CanCast(dpsSpell));
            }
        }

        public class PullSpellBalance : Action
        {
            protected override RunStatus Run(object context)
            {
                string dpsSpell = Settings.PullSpellBalance;

                if (ClassHelpers.Druid.CLCPullSpellBalance.ToUpper().Contains("AUTOMATIC")) dpsSpell = ClassHelpers.Druid.PullSpellBalance;

                Target.Face();
                bool result = Spell.Cast(dpsSpell);
                Utils.LagSleep();
                while (Me.IsCasting) System.Threading.Thread.Sleep(250);
				Spell.Cast("Moonfire");
				System.Threading.Thread.Sleep(500);
                Timers.Reset("Pull");
                

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Wrath Pull
        public class NeedToWrathPull : Decorator
        {
            public NeedToWrathPull(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                const string dpsSpell = "Wrath";

                if (Spell.IsKnown(Settings.PullSpellBalance) && !Spell.IsOnCooldown(Settings.PullSpellBalance)) return false;
                return Utils.CCheck(dpsSpell, "Timerexpired:Pull|2000");
            }
        }

        public class WrathPull : Action
        {
            protected override RunStatus Run(object context)
            {
                const string dpsSpell = "Wrath";

                Target.Face();
                System.Threading.Thread.Sleep(500);
                bool result = Spell.Cast(dpsSpell);

                if (CLC.ResultOK(Settings.MoonfireBalance))
                {
                    while (Spell.IsGCD) System.Threading.Thread.Sleep(250);
                    Spell.Cast("Moonfire");
                }

                while (Me.IsCasting) System.Threading.Thread.Sleep(500);
                Timers.Reset("Pull");

                return result ? RunStatus.Success : RunStatus.Failure;
            }
        }
        #endregion

        #region Casting So Stop Moving
        public class NeedToCastingSoStopMoving : Decorator
        {
            public NeedToCastingSoStopMoving(Composite child) : base(child) { }

            protected override bool CanRun(object context)
            {
                return Me.IsCasting;
            }
        }

        public class CastingSoStopMoving : Action
        {
            protected override RunStatus Run(object context)
            {
                return RunStatus.Success;
            }
        }
        #endregion

        #endregion
        #endregion


    }
}